#!/usr/bin/env perl

package GaiaLBS::WebService;

use Carp qw( confess );
$SIG{__DIE__} =  \&confess;
$SIG{__WARN__} = \&confess;

use utf8;
use Mojolicious::Lite;
use Mojo::ByteStream 'b';
use common::sense;
use AnyEvent;

use feature 'say';
use Data::Dumper::Simple;

my $DEBUG	= 1;
$| = 1;

app->config(hypnotoad => {listen => ['http://*:3000']});


app->ua->on(error => sub {
	my ($ua, $err) = @_;
	say "This looks bad: $err";
});
app->ua->connect_timeout(600);
app->ua->inactivity_timeout(600);
app->ua->max_connections(8196);
app->ua->request_timeout(600);


my $json = Mojo::JSON->new;

get '/' => sub {
	my $self = shift;
	#~ $self->render_later;
	#~ $self->render('index');
	$self->redirect_to('http://GaiaSup.com');
};

# query_square
get '/query/square/#apikey/#layer/#ident/#xstart/#ystart/#zstart/#xstop/#ystop/#zstop' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $apikey  = $self->param('apikey');
	my $layer  = $self->param('layer');
	my $ident  = $self->param('ident');
	my $xstart  = $self->param('xstart');
	my $ystart  = $self->param('ystart');
	my $zstart  = $self->param('zstart');
	my $xstop  = $self->param('xstop');
	my $ystop  = $self->param('ystop');
	my $zstop  = $self->param('zstop');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	my $validate_url = 'http://localhost:3001/valid/';
	my $url;
	my @error;
	
	
	my $cv_local = AE::cv;
	$cv_local->begin (sub { shift->send ( '', \@error ) });
	
	
	$url = $validate_url."$apikey/$layer/$ident";
	#~ say Dumper($url);
	$cv_local->begin;
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		my $resp = $tx->res->body;
		my $data = $json->decode( $resp );
		#~ say Dumper($resp);
		#~ say Dumper($data);
		push @error, 'node does not exist' unless $data->[0] eq 'OK';
		$cv_local->end;
	});
	
	
	$cv_local->end;
	
	$cv_local->cb( sub {
		my ($result, $err) = shift->recv;
		
		if (@$err) {
			$self->render_json( \@{['ERROR', $err]} );
		} else {
			$self->ua->get("http://localhost:3001/query_square/$apikey/$layer/$ident/$xstart/$ystart/$zstart/$xstop/$ystop/$zstop" => sub {
				my ($ua, $tx) = @_;
				$self->render_data($tx->res->body);
			});
		}
	});
	
};


# FIXME: /publish/pos/#apikey/#layer/#ident/#x/#y/#z
post '/publish/pos/#apikey/#layer/#ident' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $apikey  = $self->param('apikey');
	my $layer  = $self->param('layer');
	my $ident  = $self->param('ident');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	my $pos_json = $self->req->body;
	#~ my $pos_json = $self->req->param('pos');
	say Dumper($pos_json) if $DEBUG;
	$pos_json = b($pos_json)->encode('UTF-8');
	my $pos = $json->decode( $pos_json );
	
	say Dumper( \@{[ $self->req->param ]} ) if $DEBUG;
	
	say "GOT DATA in /publish/pos:" if $DEBUG;
	say Dumper($pos) if $DEBUG;
	
	
	my $validate_url = 'http://localhost:3001/valid/';
	my $url;
	my @error;
	
	
	my $cv_local = AE::cv;
	$cv_local->begin (sub { shift->send ( '', \@error ) });
	
	
	$url = $validate_url."$apikey/$layer/$ident";
	#~ say Dumper($url);
	$cv_local->begin;
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		my $resp = $tx->res->body;
		my $data = $json->decode( $resp );
		#~ say Dumper($resp);
		#~ say Dumper($data);
		push @error, 'node does not exist' unless $data->[0] eq 'OK';
		$cv_local->end;
	});
	
	
	$cv_local->end;
	
	
	$cv_local->cb( sub {
		my ($result, $err) = shift->recv;
		
		if (@$err) {
			$self->render_json( \@{['ERROR', $err]} );
		} else {
			$self->ua->post("http://localhost:3001/store_pos/$apikey/$layer/$ident.json" => $pos_json => sub {
				my ($ua, $tx) = @_;
				$self->render_data($tx->res->body);
			});
		}
	});
	
	# FIXME:
	# push pos to subscribers
	# 	FIXME: dispatch to gearman
	# notify neighbors
	# 	FIXME: dispatch to gearman
	# get the radius the node subscibed nearby
	# 	count the area square
	# 	query nearby nodes and push
	# publish_pos -> dispatch query_nearby -> query_node -> diff node_list, unsub previous and sub new -> check $subs_yes return via data channel
	
};


get '/query/node/#apikey/#layer/#ident' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $validate_url = 'http://localhost:3001/valid/';
	my $url;
	my @error;
	
	my $apikey	= $self->param('apikey');
	my $layer	= $self->param('layer');
	my $ident	= $self->param('ident');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	
	$url = $validate_url."$apikey/$layer/$ident";
	
	my $cv_local = AE::cv;
	$cv_local->begin (sub { shift->send ( '', \@error ) });
	
	#~ say Dumper($url);
	$cv_local->begin;
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		my $resp = $tx->res->body;
		my $data = $json->decode( $resp );
		#~ say Dumper($resp);
		#~ say Dumper($data);
		push @error, 'node does not exist' unless $data->[0] eq 'OK';
		$cv_local->end;
	});
	
	
	$cv_local->end;
	
	$cv_local->cb( sub {
		my ($result, $err) = shift->recv;
		
		if (@$err) {
			$self->render_json( \@{['ERROR', $err]} );
		} else {
			$self->ua->get("http://localhost:3001/query_node/$apikey/$layer/$ident" => sub {
				my ($ua, $tx) = @_;
				$self->render_data($tx->res->body);
			});
		}
	});
	
};


# accept multiple target via post JSON: query nodeSSSSSSSSSS
post '/query/nodes' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	my @all_nodes;
	
	
	$self->ua->on(error => sub {
		my ($ua, $err) = @_;
		say "This looks bad: $err";
	});
	$self->ua->connect_timeout(600);
	$self->ua->inactivity_timeout(600);
	$self->ua->max_connections(8196);
	$self->ua->request_timeout(600);
	Mojo::IOLoop->stream($self->tx->connection)->timeout(600);
	
	my $post_json = $self->req->body;
	#~ my $post_json = $self->req->param('post');
	say Dumper($post_json) if $DEBUG;
	$post_json = b($post_json)->encode('UTF-8');
	my $data = $json->decode( $post_json );
	
	say "GOT DATA in /query/nodes:" if $DEBUG;
	say Dumper($data) if $DEBUG;
	
	
	for my $target (@$data) {
		my @syntax_error;
		return $self->render_json( \@{['', \@syntax_error]} )
			if @syntax_error
				= &_syntax_error(
					apikey => $target->{'apikey'},
					layer => $target->{'layer'}
				);
	}
	
	
	my $validate_url = 'http://localhost:3001/valid/';
	my $url;
	my @error;
	
	
	my $cv_local = AE::cv;
	$cv_local->begin (sub { shift->send ( \@all_nodes, \@error ) });
	
	
	push @error, 'no target to query' unless @$data;
	
	for my $target (@$data) {
		my $target_apikey	= $target->{'apikey'};
		my $target_layer	= $target->{'layer'};
		my $target_ident	= $target->{'ident'};
		
		$url = $validate_url."$target_apikey/$target_layer/$target_ident";
		#~ say Dumper($url);
		$cv_local->begin;
		$self->ua->get($url => sub {
			my ($ua, $tx) = @_;
			my $resp = $tx->res->body;
			my $data = $json->decode( $resp );
			say Dumper($resp) if $DEBUG;
			#~ say Dumper($data);
			push @error, "target does not exist: apikey => $target_apikey, layer => $target_layer, ident => $target_ident" unless $data->[0] eq 'OK';
			$cv_local->end;
		});
	}
	
	
	$cv_local->end;
	
	
	my $cv_valid = AE::cv;
	
	$cv_local->cb( sub {
		my ($result, $err) = shift->recv;
		
		say Dumper( $result, $err ) if @$err;
		
		
		if (@$err) {
			$cv_valid->send( 'ERROR', $err );
		} else {
			$cv_valid->begin (sub { shift->send ( \@all_nodes, \@error ) });
			for my $target (@$data) {
				my $target_apikey	= $target->{'apikey'};
				my $target_layer	= $target->{'layer'};
				my $target_ident	= $target->{'ident'};
				$cv_valid->begin;
				$self->ua->get("http://localhost:3001/query_node/$target_apikey/$target_layer/$target_ident" => sub {
					my ($ua, $tx) = @_;
					my $resp = $tx->res->body;
					my $data = $json->decode( $resp );
					say Dumper($resp) if $DEBUG;
					say Dumper($data);
					my ($data_node, $err) = @$data;
					say Dumper($data_node, $err);
					
					@$err ?
						push @error, "query error: apikey => $target_apikey, layer => $target_layer, ident => $target_ident"
					:
						push @all_nodes, $data->[0]
					;
					
					$cv_valid->end;
				});
			}
			$cv_valid->end;
		}
	});
	
	
	
	$cv_valid->cb( sub {
		my ($result, $err) = shift->recv;
		
		say 'Final: '.Dumper( $result, $err ) if $DEBUG;
		
		if (@$err) {
			$self->render_json( \@{['ERROR', $err]} );
		} else {
			$self->render_json( \@{[$result, $err]} );
		}
	});
	
};


# accept multiple target via post JSON: subscribe nodeSSSSSSSSSS
post '/subscribe/nodes/by/#apikey/#layer/#ident' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	$self->ua->on(error => sub {
		my ($ua, $err) = @_;
		say "This looks bad: $err";
	});
	$self->ua->connect_timeout(600);
	$self->ua->inactivity_timeout(600);
	$self->ua->max_connections(8196);
	$self->ua->request_timeout(600);
	Mojo::IOLoop->stream($self->tx->connection)->timeout(600);
	
	
	my $post_json = $self->req->body;
	#~ my $post_json = $self->req->param('post');
	say Dumper($post_json) if $DEBUG;
	$post_json = b($post_json)->encode('UTF-8');
	my $data = $json->decode( $post_json );
	
	say "GOT DATA in /subscribe/nodes:" if $DEBUG;
	say Dumper($data) if $DEBUG;
	
	my $apikey	= $self->param('apikey');
	my $layer	= $self->param('layer');
	my $ident	= $self->param('ident');
	
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	for my $target (@$data) {
		my @syntax_error;
		return $self->render_json( \@{['', \@syntax_error]} )
			if @syntax_error = &_syntax_error( apikey => $target->{'apikey'}, layer => $target->{'apikey'} );
	}
	
	
	my $validate_url = 'http://localhost:3001/valid/';
	my $url;
	my @error;
	
	
	my $cv_local = AE::cv;
	$cv_local->begin (sub { shift->send ( '', \@error ) });
	
	$url = $validate_url."$apikey/$layer/$ident";
	#~ say Dumper($url);
	$cv_local->begin;
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		my $resp = $tx->res->body;
		my $data = $json->decode( $resp );
		say Dumper($resp) if $DEBUG;
		#~ say Dumper($data);
		push @error, "subscriber does not exist: apikey => $apikey, layer => $layer, ident => $ident" unless $data->[0] eq 'OK';
		$cv_local->end;
	});
	
	push @error, 'no target to subscribe' unless @$data;
	
	for my $target (@$data) {
		my $target_apikey	= $target->{'apikey'};
		my $target_layer	= $target->{'layer'};
		my $target_ident	= $target->{'ident'};
		
		$url = $validate_url."$target_apikey/$target_layer/$target_ident";
		#~ say Dumper($url);
		$cv_local->begin;
		$self->ua->get($url => sub {
			my ($ua, $tx) = @_;
			my $resp = $tx->res->body;
			my $data = $json->decode( $resp );
			say Dumper($resp) if $DEBUG;
			#~ say Dumper($data);
			push @error, "target does not exist: apikey => $target_apikey, layer => $target_layer, ident => $target_ident" unless $data->[0] eq 'OK';
			$cv_local->end;
		});
	}
	
	
	$cv_local->end;
	
	
	my $cv_valid = AE::cv;
	
	$cv_local->cb( sub {
		my ($result, $err) = shift->recv;
		
		say Dumper( $result, $err ) if @$err;
		
		
		if (@$err) {
			$cv_valid->send( 'ERROR', $err );
		} else {
			$cv_valid->begin (sub { shift->send ( '', \@error ) });
			for my $target (@$data) {
				my $target_apikey	= $target->{'apikey'};
				my $target_layer	= $target->{'layer'};
				my $target_ident	= $target->{'ident'};
				$cv_valid->begin;
				$self->ua->get("http://localhost:3001/stalk/$target_apikey/$target_layer/$target_ident/by/$apikey/$layer/$ident" => sub {
					my ($ua, $tx) = @_;
					my $resp = $tx->res->body;
					my $data = $json->decode( $resp );
					say Dumper($resp) if $DEBUG;
					#~ say Dumper($data);
					push @error, "subscribe error: apikey => $target_apikey, layer => $target_layer, ident => $target_ident" unless $data->[0] eq 'OK';
					
					$cv_valid->end;
				});
			}
			$cv_valid->end;
		}
	});
	
	
	
	$cv_valid->cb( sub {
		my ($result, $err) = shift->recv;
		
		say 'Final: '.Dumper( $result, $err ) if $DEBUG;
		
		if (@$err) {
			$self->render_json( \@{['ERROR', $err]} );
		} else {
			$self->render_json( \@{['OK', $err]} );
		}
	});
	
};


# register layer
get '/register/#apikey/#layer' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $register_url = 'http://localhost:3001/register/';
	my $url;
	my $apikey	= $self->param('apikey');
	my $layer	= $self->param('layer');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	$url = $register_url."$apikey/$layer";
	
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		$self->render_data($tx->res->body);
	});
};


# register node
get '/register/#apikey/#layer/#ident' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $register_url = 'http://localhost:3001/register/';
	my $url;
	my $apikey	= $self->param('apikey');
	my $layer	= $self->param('layer');
	my $ident	= $self->param('ident');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	$url = $register_url."$apikey/$layer/$ident";
	
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		$self->render_data($tx->res->body);
	});
};


# revoke layer
get '/revoke/#apikey/#layer' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $revoke_url = 'http://localhost:3001/revoke/';
	my $url;
	my $apikey	= $self->param('apikey');
	my $layer	= $self->param('layer');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	$url = $revoke_url."$apikey/$layer";
	
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		$self->render_data($tx->res->body);
	});
};


# revoke node
get '/revoke/#apikey/#layer/#ident' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $revoke_url = 'http://localhost:3001/revoke/';
	my $url;
	my $apikey	= $self->param('apikey');
	my $layer	= $self->param('layer');
	my $ident	= $self->param('ident');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	$url = $revoke_url."$apikey/$layer/$ident";
	
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		$self->render_data($tx->res->body);
	});
};


get '/valid/#apikey/#layer/#ident' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $validate_url = 'http://localhost:3001/valid/';
	my $url;
	my $apikey	= $self->param('apikey');
	my $layer	= $self->param('layer');
	my $ident	= $self->param('ident');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	$url = $validate_url."$apikey/$layer/$ident";
	
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		$self->render_data($tx->res->body);
	});
	
};


get '/subscribers/#apikey/#layer/#ident' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $validate_url = 'http://localhost:3001/valid/';
	my $url;
	my @error;
	
	
	my $apikey	= $self->param('apikey');
	my $layer	= $self->param('layer');
	my $ident	= $self->param('ident');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	
	my $cv_local = AE::cv;
	$cv_local->begin (sub { shift->send ( '', \@error ) });
	
	$url = $validate_url."$apikey/$layer/$ident";
	#~ say Dumper($url);
	$cv_local->begin;
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		my $resp = $tx->res->body;
		my $data = $json->decode( $resp );
		#~ say Dumper($resp);
		#~ say Dumper($data);
		push @error, 'node does not exist' unless $data->[0] eq 'OK';
		$cv_local->end;
	});
	
	
	$cv_local->end;
	
	$cv_local->cb( sub {
		my ($result, $err) = shift->recv;
		
		if (@$err) {
			$self->render_json( \@{['ERROR', $err]} );
		} else {
			$self->ua->get("http://localhost:3001/stalkers/$apikey/$layer/$ident" => sub {
				my ($ua, $tx) = @_;
				$self->render_data($tx->res->body);
			});
		}
	});
	
};


get '/subscribe/node/#target_apikey/#target_layer/#target_ident/by/#apikey/#layer/#ident' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $validate_url = 'http://localhost:3001/valid/';
	my $url;
	my @error;
	
	
	my $target_apikey	= $self->param('target_apikey');
	my $target_layer	= $self->param('target_layer');
	my $target_ident	= $self->param('target_ident');
	
	my $apikey	= $self->param('apikey');
	my $layer	= $self->param('layer');
	my $ident	= $self->param('ident');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $target_apikey, layer => $target_layer );
	
	my $cv_local = AE::cv;
	$cv_local->begin (sub { shift->send ( '', \@error ) });
	
	
	$url = $validate_url."$target_apikey/$target_layer/$target_ident";
	#~ say Dumper($url);
	$cv_local->begin;
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		my $resp = $tx->res->body;
		my $data = $json->decode( $resp );
		#~ say Dumper($resp);
		#~ say Dumper($data);
		push @error, 'target does not exist' unless $data->[0] eq 'OK';
		$cv_local->end;
	});
	
	
	$url = $validate_url."$apikey/$layer/$ident";
	#~ say Dumper($url);
	$cv_local->begin;
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		my $resp = $tx->res->body;
		my $data = $json->decode( $resp );
		#~ say Dumper($resp);
		#~ say Dumper($data);
		push @error, 'subscriber does not exist' unless $data->[0] eq 'OK';
		$cv_local->end;
	});
	
	
	$cv_local->end;
	
	$cv_local->cb( sub {
		my ($result, $err) = shift->recv;
		
		if (@$err) {
			$self->render_json( \@{['ERROR', $err]} );
		} else {
			$self->ua->get("http://localhost:3001/stalk/$target_apikey/$target_layer/$target_ident/by/$apikey/$layer/$ident" => sub {
				my ($ua, $tx) = @_;
				$self->render_data($tx->res->body);
			});
		}
	});
	
};


any '/#whatever' => sub {
	my $self = shift;
	#~ $self->render_later;
	#~ $self->render('index');
	$self->redirect('http://GaiaSup.com');
};


# Message


post '/msg/send/#apikey/#layer/#ident' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $apikey  = $self->param('apikey');
	my $layer  = $self->param('layer');
	my $ident  = $self->param('ident');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	my $pos_json = $self->req->body;
	#~ my $pos_json = $self->req->param('pos');
	say Dumper($pos_json) if $DEBUG;
	#~ $pos_json = b($pos_json)->decode('UTF-8');
	#~ say Dumper($pos_json) if $DEBUG;
	my $pos = $json->decode( $pos_json );
	
	say "GOT DATA in /msg/send:" if $DEBUG;
	say Dumper($pos) if $DEBUG;
	
	
	my $validate_url = 'http://localhost:3001/valid/';
	my $url;
	my @error;
	
	
	my $cv_local = AE::cv;
	$cv_local->begin (sub { shift->send ( '', \@error ) });
	
	
	$url = $validate_url."$apikey/$layer/$ident";
	#~ say Dumper($url);
	$cv_local->begin;
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		my $resp = $tx->res->body;
		my $data = $json->decode( $resp );
		#~ say Dumper($resp);
		#~ say Dumper($data);
		push @error, 'node does not exist' unless $data->[0] eq 'OK';
		$cv_local->end;
	});
	
	
	$cv_local->end;
	
	
	$cv_local->cb( sub {
		my ($result, $err) = shift->recv;
		
		if (@$err) {
			$self->render_json( \@{['ERROR', $err]} );
		} else {
			$self->ua->post("http://localhost:3001/msg/send/$apikey/$layer/$ident" => $pos_json => sub {
				my ($ua, $tx) = @_;
				$self->render_data($tx->res->body);
			});
		}
	});
	
};


get '/msg/query/#apikey/#layer/#ident/#start/#stop' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $validate_url = 'http://localhost:3001/valid/';
	my $url;
	my @error;
	
	
	my $apikey	= $self->param('apikey');
	my $layer	= $self->param('layer');
	my $ident	= $self->param('ident');
	my $start	= $self->param('start');
	my $stop	= $self->param('stop');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	
	my $cv_local = AE::cv;
	$cv_local->begin (sub { shift->send ( '', \@error ) });
	
	$url = $validate_url."$apikey/$layer/$ident";
	#~ say Dumper($url);
	$cv_local->begin;
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		my $resp = $tx->res->body;
		my $data = $json->decode( $resp );
		#~ say Dumper($resp);
		#~ say Dumper($data);
		push @error, 'node does not exist' unless $data->[0] eq 'OK';
		$cv_local->end;
	});
	
	
	$cv_local->end;
	
	$cv_local->cb( sub {
		my ($result, $err) = shift->recv;
		
		if (@$err) {
			$self->render_json( \@{['ERROR', $err]} );
		} else {
			$self->ua->get("http://localhost:3001/msg/query/$apikey/$layer/$ident/$start/$stop" => sub {
				my ($ua, $tx) = @_;
				$self->render_data($tx->res->body);
			});
		}
	});
	
};


get '/msg/flush/#apikey/#layer/#ident' => sub {
	my $self = shift;
	$self->res->headers->header('Access-Control-Allow-Origin' => '*');
	$self->render_later;
	
	my $validate_url = 'http://localhost:3001/valid/';
	my $url;
	my @error;
	
	
	my $apikey	= $self->param('apikey');
	my $layer	= $self->param('layer');
	my $ident	= $self->param('ident');
	
	my @syntax_error;
	return $self->render_json( \@{['', \@syntax_error]} )
		if @syntax_error = &_syntax_error( apikey => $apikey, layer => $layer );
	
	
	my $cv_local = AE::cv;
	$cv_local->begin (sub { shift->send ( '', \@error ) });
	
	$url = $validate_url."$apikey/$layer/$ident";
	#~ say Dumper($url);
	$cv_local->begin;
	$self->ua->get($url => sub {
		my ($ua, $tx) = @_;
		my $resp = $tx->res->body;
		my $data = $json->decode( $resp );
		#~ say Dumper($resp);
		#~ say Dumper($data);
		push @error, 'node does not exist' unless $data->[0] eq 'OK';
		$cv_local->end;
	});
	
	
	$cv_local->end;
	
	$cv_local->cb( sub {
		my ($result, $err) = shift->recv;
		
		if (@$err) {
			$self->render_json( \@{['ERROR', $err]} );
		} else {
			$self->ua->get("http://localhost:3001/msg/flush/$apikey/$layer/$ident" => sub {
				my ($ua, $tx) = @_;
				$self->render_data($tx->res->body);
			});
		}
	});
	
};




# Internal function

sub _syntax_error {
	my @error;
	
	say Dumper(@_) if $DEBUG;
	
	push @error, 'No data' unless @_;
	
	my %args = @_;
	say Dumper(%args) if $DEBUG;
	
	# check user id
	if ( $args{'id'} and $args{'id'} =~ /[\/\?#=@]/ ) {
		push @error, 'User id syntax error';
	}
	
	# check userhash
	if ($args{'userhash'} and ( $args{'userhash'} =~ /\W/ or length( $args{'userhash'} ) != 40 )) {
		push @error, 'Invalid user';
	}
	
	# check apikey
	if ($args{'apikey'} and ( $args{'apikey'} =~ /\W/ or length( $args{'apikey'} ) != 40 )) {
		push @error, 'Invalid apikey';
	}
	
	# check email
	if ($args{'email'} and not $args{'email'} =~ /^[\w\.]+\@(\w+\.)*[a-z]{2,5}$/i) {
		push @error, 'Invalid email';
	}
	
	# check passwd hash
	if ($args{'passwd'} and ( (not $args{'passwd'} =~ /[a-z0-9]/i) or length( $args{'passwd'} ) != 128 )) {
		push @error, 'Invalid passwd';
	}
	
	say Dumper(@error) if $DEBUG;
	return @error ? @error : () ;
}


app->start;



__DATA__

@@ layouts/default.html.ep
<!DOCTYPE html>
<html>
  <head><title><%= title %></title></head>
  <body><%= content %></body>
</html>

@@ index.html.ep
% layout 'default';
% title 'Welcome';
Welcome to the Mojolicious real-time web framework!

@@ publish_pos.html.ep
% layout 'default';
% title 'Welcome';
%= form_for '' => (method => 'post') => begin
	loc:x：<br />
	<%= text_field 'x' => 'x' %><br />
	loc:y：<br />
	<%= text_field 'y' => 'y' %><br />
	loc:z：<br />
	<%= text_field 'z' => 'z' %><br />
	move：<br />
	<%= text_field 'move_x1' => 'move_x1' %><%= text_field 'move_y1' => 'move_y1' %><%= text_field 'move_z1' => 'move_z1' %><br />
	<%= text_field 'move_x2' => 'move_x2' %><%= text_field 'move_y2' => 'move_y2' %><%= text_field 'move_z2' => 'move_z2' %><br />
	<%= text_field 'speed' => 'speed' %><br />
	orientation：<br />
	<%= text_field 'orien1' => 'orien1' %><%= text_field 'orien2' => 'orien2' %><br />
	
	%= submit_button 'publish update'
% end

@@ subscribe_node.html.ep
% layout 'default';
% title 'Welcome';
%= form_for '' => (method => 'post') => begin
	my apikey：<br />
	<%= text_field 'apikey' => 'apikey' %><br />
	my layer：<br />
	<%= text_field 'layer' => 'layer' %><br />
	my node ident：<br />
	<%= text_field 'ident' => 'ident' %><br />
	
	%= submit_button 'publish update'
% end
